"""

VCard - Business card object manager

Contacts are stored as a stream of VCard objects.
This makes them easy to manipulate and migrate to
or from other apps.

Copyright 2001/2002 James Weatherall

"""

import string, time

from VLowLevel import VObject, VProps, VReadObjects, _VRegisterHandler

# - ADR parameter tags
Address_TYPE = (["DOM", "INTL", "PARCEL", "POSTAL", "HOME", "WORK"],
                ["INTL", "POSTAL", "PARCEL", "WORK"])

class Address:
    def __init__(s):
        s.postbox=""
        s.extended=""
        s.street=""
        s.city=""
        s.region=""
        s.postcode=""
        s.country=""
        s.comment=""
        s.types=[]
    class Property(VProps.List):
        param_types = {"TYPE":Address_TYPE}
        param_types.update(VProps.List.param_types)
        def get(self):
            a=Address()
            (a.postbox, a.extended, a.street,
             a.city, a.region, a.postcode, a.country) = VProps.List.get(self)
            a.types=self.get_param("TYPE")
            return a
        def set(self, a):
            VProps.List.set(self, [a.postbox, a.extended, a.street,
                a.city, a.region, a.postcode, a.country])
            self.set_param("TYPE", a.types)
    def __str__(s):
        v=[]
        for a in ["postbox", "extended", "street", "city",
                  "region", "postcode", "country", "comment"]:
            if getattr(s, a):
                v.append(getattr(s, a))
        return string.join(v, ",\n")

class Name:
    def __init__(s):
        s.last=""
        s.middle=""
        s.first=""
        s.title=""
    class Property(VProps.List):
        def get(s):
            n=Name()
            names=VProps.List.get(s)
            fields=["last", "first", "middle", "title"]
            for i in range(len(fields)):
                if i < len(names): setattr(n, fields[i], names[i])
                else: setattr(n, fields[i], "")
            return n
        def set(s, n):
            VProps.List.set(s, [n.last, n.first, n.middle, n.title])
    def __str__(s):
        return string.join(filter(lambda x:x, [s.last, s.middle, s.first, s.title]),
                           ", ")

Telephone_TYPE = (["PREF", "HOME", "WORK", "VOICE", "FAX", "MSG", "CELL",
                   "PAGER", "BBS", "MODEM", "CAR", "ISDN", "VIDEO"],
                  ["VOICE"])

class Telephone:
    def __init__(s):
        s.types=[]
        s.number=""
    class Property(VObject.Property):
        param_types = {"TYPE":Telephone_TYPE}
        param_types.update(VObject.Property.param_types)
        def get(s):
            n=Telephone()
            n.number = VObject.Property.get(s)
            n.types=s.get_param("TYPE")
            return n
        def set(s, n):
            VObject.Property.set(s, n.number)
            s.set_param("TYPE", n.types)
    def __str__(s):
        return s.number

Email_TYPE = (["AOL", "APPLELINK", "ATTMAIL", "CIS", "EWORLD",
               "INTERNET", "IBMMAIL", "MCIMAIL", "POWERSHARE",
               "PRODIGY", "TLX", "X400"], ["INTERNET"])

class Email:
    def __init__(s):
        s.types=[]
        s.address=""
    class Property(VObject.Property):
        param_types = {"TYPE":Email_TYPE}
        param_types.update(VObject.Property.param_types)
        def get(s):
            n=Email()
            n.address = VObject.Property.get(s)
            n.types=s.get_param("TYPE")
            return n
        def set(s, n):
            VObject.Property.set(s, n.address)
            s.set_param("TYPE", n.types)
    def __str__(s):
        return s.address

class Date:
    def __init__(s):
        pass

class Revision:
    class Property(VObject.Property):
        def get(self):
            rev=VObject.Property.get(self)[:15]
            return time.strptime(rev, "%Y%m%dT%H%M%S")
        def set(self, l):
            revstr=time.strftime("%Y%m%dT%H%M%S", time.localtime(l))
            VObject.Property.set(self, revstr)

class Photo:
    def __init__(s):
        s.photo_data=""
        s.type=""
        s.is_url=1
    class Property(VObject.Property):
        def get(s):
            p=Photo()
            p.photo_data = VObject.Property.get(s)
            types = s.get_param("TYPE")
            if not types: s.type = "JPEG"
            else: s.type = types[0]
            p.is_url = "URL" in s.get_param("VALUE")
            return p
        def set(s, p):
            VObject.Property.set(s, p.photo_data)
            s.set_param("TYPE", [p.type])
            if p.is_url: s.set_param("VALUE", ["URL"])

class VCard(VObject):
    """ VCard object class """
    # - Override VCard attributes
    typename = "VCARD"
    prop_types = {"FN":VProps.String,
                  "N":Name.Property,
                  "PHOTO":Photo.Property,
                  "BDAY":VObject.Property,
                  "NICKNAME":VProps.String,
                  "ADR":Address.Property,
                  "LABEL":VProps.String,
                  "TEL":Telephone.Property,
                  "EMAIL":Email.Property,

                  "NOTE":VProps.String,
                  "REV":Revision.Property,

                  "UID":VProps.String,
                  }

    # - The version of spec we match
    Version = 2.1

    def __init__(self, parent=None):
        VObject.__init__(self, parent, VCard.typename)
        version = VObject.Property("VERSION", "2.1")
        self.make_group(None).add_property(version)

    def handle_VERSION(self, group, params, value):
        if float(value) != VCard.Version:
            print "warning: VCard version mismatch"

    # - Ungrouped attributes

    def __get_prop(self, name, default):
        prop=self.get_group(None).get_property(name)
        if prop: return prop.get()
        return default
    def __set_prop(self, name, val):
        self.get_group(None).make_property(name, val)
    
    # - FN - formatted name
    def get_fullname(self): return self.__get_prop("FN", "")
    def set_fullname(self,name): self.__set_prop("FN", name)

    # - N - name
    def get_name(self): return self.__get_prop("N", Name())
    def set_name(self,name): self.__set_prop("N", name)

    # - BDAY - birth date
    def get_birthdate(self):
        return self.__get_prop("BDAY", Date())
    def set_birthdate(self, date):
        self.__set_prop("BDAY", date)

    # - ADR - address data
    def get_addresses(self):
        """ get_addresses()
            locate all addresses in all groups
        """
        props = self.get_properties("ADR")
        return map(lambda p:p.get(), props)

    # - PHOTO - photograph data
    def get_photo_jpeg_url(self):
        photo = self.__get_prop("PHOTO", "")
        if not photo.is_url: return ""
        return photo.photo_data
    def set_photo_jpeg_url(self, url):
        photo=Photo()
        photo.photo_data=url
        photo.is_url=1
        photo.type="JPEG"
        self.__set_prop("PHOTO", photo)

    # - Single-address per-type handlers
    #   If there are multiple addresses matching a single specified type
    #   then the first one is used & replaced.
    def _get_address(self, types):
        types=map(lambda t:("TYPE", t), types)
        prop = self.get_group(None).get_property("ADR", types)
        if prop: return prop.get()
        return Address()
    def _set_address(self, types, addr):
        if not str(addr): addr=None
        types=map(lambda t:("TYPE", t), types)
        self.get_group(None).make_property("ADR", addr, types)
        # - For safety, clear out all "LABEL" attributes
        props = self.get_group(None).get_properties("LABEL")
        for prop in props:
            self.rem_property(prop)
        
    def get_home_address(self): return self._get_address(["HOME"])
    def get_work_address(self): return self._get_address(["WORK"])
    def set_work_address(self, addr): # - Default
        addr.types = []
        self._set_address(["WORK"], addr)
    def set_home_address(self, addr):
        addr.types = ["INTL", "POSTAL", "PARCEL", "HOME"]
        self._set_address(["HOME"], addr)

    # - TEL - telephone data
    def get_phone_nos(self):
        """ get_phone_nos()
            locate all phone nos in all groups
        """
        addrs=[]
        props = self.get_properties("TEL")
        for prop in props:
            addrs.append(prop.get())
        return addrs

    def _get_phone(self, types):
        types=map(lambda t:("TYPE", t), types)
        prop = self.get_group(None).get_property("TEL", types)
        if prop: return prop.get()
        return Telephone()
    def _set_phone(self, types, addr):
        if not addr.number: addr=None
        else: addr.types = types
        types=map(lambda t:("TYPE", t), types)
        prop=self.get_group(None).make_property("TEL", addr, types)

    def get_home_phone(self): return self._get_phone(["VOICE", "HOME"])
    def get_work_phone(self): return self._get_phone(["VOICE", "WORK"])
    def get_mobile_phone(self): return self._get_phone(["VOICE", "CELL"])
    def set_home_phone(self, p): self._set_phone(["VOICE", "HOME"], p)
    def set_work_phone(self, p): self._set_phone(["VOICE", "WORK"], p)
    def set_mobile_phone(self, p): self._set_phone(["VOICE", "CELL"], p)

    # - EMAIL - email addresses
    def get_email_addrs(self):
        """ get all the INTERNET email addresses in all groups """
        props = self.get_properties("EMAIL", [("TYPE", "INTERNET")])
        return map(lambda x:x.get(), props)
    def set_email_addrs(self, addrs):
        """ set the INTERNET email addresses """
        props = self.get_properties("EMAIL", [("TYPE", "INTERNET")])
        for prop in props:
            if prop.get() in addrs:
                addrs.remove(prop.get())
            else:
                self.rem_property(prop)
        for addr in addrs:
            prop=self._create_property("EMAIL")
            prop.set(addr)
            self.add_property(prop)

    # - NOTE - comment about this VCard
    def get_note(self): return self.__get_prop("NOTE", "")
    def set_note(self, value): self.__set_prop("NOTE", value)

    # - REV - revision of this VCard
    def get_revision(self): return self.__get_prop("REV", 0)
    def set_revision(self): self.__set_prop("REV", time.time())

    # ***

    # - NICKNAME - nickname
    def get_nickname(self):
        return self.__get_prop("NICKNAME", "")
    def set_nickname(self,name):
        self.__set_prop("NICKNAME", name)

    # - UID - unique identifier
    def get_unique_id(self):
        return self.__get_prop("UID", "")

# - Register with VLowLevel
_VRegisterHandler(VCard)

if __name__ == "__main__":
    f=open("D:\\james\\Python\\Jen Yee Chan.txt")
    card = VReadObjects(f)[0]
    print "Fullname:", card.get_fullname()
    print "Name:", card.get_name()
    print "Nickname:", card.get_nickname()
    print "Addresses:"
    for addr in card.get_addresses():
        print addr
