"""

VLowLevel - Low level handling of VCard/VCal objects

Reads a stream and attempts to read each line as a
property.  For each successfully read property, the
engine invokes the appropriate handle_<propname>
method.

Copyright 2001/2002 James Weatherall

"""

import string
import re

#
# -=- LOW LEVEL VCard/VCal..etc reading/writing classes
#

def _handle(object, name, params):
    mn="handle_"+name
    if hasattr(object, mn):
	return apply(getattr(object, mn), params)
    elif hasattr(object, "handle_unknown"):
	return apply(getattr(object, "handle_unknown"), (name,)+params)

class VReader:
    """ Core class used to read in data, line by line """

    _RE_Property = re.compile("^(?:([^.]+)\.)*([^;:]+)(?:;([^:]+))?:([^\r\n]*)")

    def __init__(self, stream, handler):
        if type(stream) == type(""):
            import StringIO
            stream = StringIO.StringIO(stream)
        self.stream = stream
        self.__linecache = None

	while 1:
	    property = self.__readProperty()
	    if not property:
	        _handle(handler, "eof", ())
	        break

	    # - Process the property
	    _handle(handler, property[0], property[1:])

    def __readline(self):
        """ readline() - reads a line from the stream.
            Unfolds folded lines if required. """
        if self.__linecache:
            line=self.__linecache
            self.__linecache = None
	else:
            line = ""
        while 1:
            data = self.stream.readline()
            if not data: break
            data = string.rstrip(data)
            if line:
                if data and (data[0] in string.whitespace):
                    line=line+data
                else:
                    self.__linecache=data
                    break
            else:
                line=data
        return line

    def __readProperty(self):
        """ readProperty() - read a single property. """
	while 1:
            line = self.__readline()
            if not line: return None
            result = self._RE_Property.match(line)
            if not result:
                print "unmatched line:", line
                continue
            # *** We don't presently handle the group attribute
            group, name, paramstxt, value= result.groups()
            if group: group = string.upper(group)
            name = string.upper(name)
            params=[]
            if paramstxt:
            	paramslist=string.split(paramstxt, ";")
                for item in paramslist:
                    item = string.upper(item)
                    bits = string.split(item, "=")
                    if len(bits) == 2:
                        params.append((bits[0], bits[1]))
                    else:
                        params.append((None,item))
            return name, group, params, value

class VHandler:
    def __init__(self):
        pass

#
# -=- HIGH LEVEL reader/writer classes!
#

_ObjectTypes = {}

_Property_ENCODING = (["7BIT", "8BIT", "QUOTED-PRINTABLE", "BASE64"], ["7BIT"])
_Property_VALUE = (["URL", "CONTENT-ID"], [])

class VObject:
    typename = ""
    prop_types={}
    parent = None

    # - Properties
    class Property:
        param_types = {"ENCODING": _Property_ENCODING,
                       "VALUE": _Property_VALUE}
        def __init__(self, name, value, params=[], group=None):
            self.name = name
            self.group = group
            self._rawparams = []
            self._params = {}
            self.value = value
            self.object = None
            while params:
                pname, pval=params.pop()
                if not pname:
                    for ptype in self.param_types.keys():
                        if pval in self.param_types[ptype][0]:
                            pname = ptype
                            break
                if pname:
                    if not self._params.has_key(pname):
                        self._params[pname] = []
                    if pval not in self._params[pname]:
                        self._params[pname].append(pval)
                else:
                    self._rawparams.append((pname, pval))
            for ptype in self.param_types.keys():
                if not self._params.has_key(ptype):
                    self._params[ptype] = self.param_types[ptype][1][:]
        def __repr__(self):
            if self.group:v=self.group+"."
            else: v=""
            v=v+self.name
            for k in self._params.keys():
                if (self.param_types.has_key(k) and
                    (self._params[k] != self.param_types[k][1])):
                    # - Only put in non-default settings
                    for p in self._params[k]:
                        v=v+";"
                        # - If parameter name is ambiguous, insert it!
                        for kk in self.param_types.keys():
                            if (p in self.param_types[kk][0]) and (kk != k):
                                v=v+k+"="
                                break
                        v=v+p
            for k in self._rawparams:
                v=v+";"
                if k[0]:v=v+k[0]+"="
                v=v+k[1]
            v=v+":"+self.value
            return v

        # - Parameters
        def get_params(self):
            return self._params
        def get_param(self, name):
            if self._params.has_key(name):
                return self._params[name]
            else:
                v=[]
                for pname, pval in self._rawparams:
                    if name == pname:
                        v.append(pval)
                return v
        def set_param(self, name, value):
            if not value and self.param_types.has_key(name):
                value = self.param_types[name][1]
            if value: self._params[name]=value[:]
        def has_param(self, name, value):
            return value in self.get_param(name)

        # - Fetching the value and decoding it
        def get(self):
            # - Unencode the data if required
            enc=self.get_param("ENCODING")
            if "BASE64" in enc:
                import base64
                return base64.decodestring(self.value)
            elif "QUOTED-PRINTABLE" in enc:
                sout=""
                sin=self.value
                while sin:
                    try:
                        i=string.index(sin, "=")
                    except:
                        sout=sout+sin
                        break
                    sout=sout+sin[:i]
                    c=chr(eval("0x"+sin[i+1:i+3]))
                    sout=sout+c
                    sin=sin[i+3:]
                return sout
            elif ("7BIT" in enc) or ("8BIT" in enc):
                return self.value
        def set(self, value):
            # - If the data contains funny characters then
            #   let's use a different encoding...
            enc=self.param_types["ENCODING"][1][0]
            for c in ["\000"]:
                if c in value: enc="BASE64"
            if enc!="BASE64":
                for c in ["\n","\r"]:
                    if c in value: enc="QUOTED_PRINTABLE"
            self.set_param("ENCODING", [enc])
            if enc=="BASE64":
                import base64
                self.value = base64.encodestring(value)
            elif enc=="QUOTED-PRINTABLE":
                sout=""
                for c in value:
                    if c in "\n\r=":
                        sout=sout+string.zfill(hex(ord(c))[2:], 2)
                    else:
                        sout=sout+c
                self.value=sout
            elif enc in ["7BIT", "8BIT"]:
                self.value=value

        # - Finding properties in the same group    
        def get_properties(self, name, params=[]):
            if self.object:
                return self.object.get_group(self.group).get_properties(name, params)
            else: return []
        def get_property(self, name, params=[]):
            if self.object:
                return self.object.get_group(self.group).get_property(name, params)
            else: return None

    # - Property groups
    class Group:
        def __init__(self, name, object):
            self.name = name
            self.object = object
            self.properties = []
        def __repr__(self):
            v=""
            for property in self.properties:
                v=v+str(property)+"\r\n"
            return v

        # - Basic property manipulation
        def add_property(self, prop):
            self.properties.append(prop)
            prop.group = self.name
            prop.object = self.object
        def rem_property(self, prop):
            if prop in self.properties:
                self.properties.remove(prop)
                prop.object=None

        # - More advanced manipulations
        def get_properties(self, name, params=[]):
            result = []
            for property in self.properties:
                if property.name == name:
                    ok=1
                    for param in params:
                        if not property.has_param(param[0], param[1]):
                            ok=0
                            break
                    if ok: result.append(property)
            return result
        def get_property(self, name, params=[]):
            p=self.get_properties(name, params)
            if len(p) > 1: print "warning: more than one match"
            if p: return p[0]
            else: return None
        def make_property(self, name, value, params=[]):
            p = self.get_property(name, params)
            if not value:
                self.rem_property(p)
                return None
            if p:
                p.set(value)
                p.params = params
            else:
                p = self.object._create_property(name, "", params, self.name)
                p.set(value)
                self.add_property(p)
            return p

    # - Properties within an object are unique
    #   if the combination of group, name and parameters
    #   is unique
    
    # - Base methods
    def __init__(self, parent, typename=None):
        self.groups = {}
        self.children = []
        if self.typename != typename: self.typename=typename
        if self.parent != parent: self.parent=parent

    # - Handlers
    def handle_subobject(self, object):
        self.children.append(object)
    def handle_unknown(self, name, group, params, value):
        self.add_property(self._create_property(name, value, params, group))

    def _create_property(self, name, value="", params=[], group=None):
        if self.prop_types.has_key(name):
            prop=self.prop_types[name](name, value, params, group)
        else:
            prop=self.Property(name, value, params, group)
        return prop

    # - Group handling
    def get_groups(self):
        return self.groups.keys()
    def get_group(self, group):
        return self.groups[group]
    def make_group(self, group):
        if not self.groups.has_key(group):
            self.groups[group]=VObject.Group(group, self)
        return self.groups[group]

    # - Property handling - accesses ALL groups!
    def add_property(self, prop):
        self.make_group(prop.group).add_property(prop)
    def rem_property(self, prop):
        self.get_group(prop.group).rem_property(prop)

    def get_properties(self, name, params=[]):
        result = []
        for group in self.groups.values():
            result = result + group.get_properties(name, params)
        return result
    def make_property(self, name, value, params=[], group=None):
        return self.make_group(group).make_property(name, value, params)

    def __repr__(self):
        v=(["BEGIN:%s\r\n"%self.typename]+
           map(str, self.groups.values())+
           map(str, self.children)+
           ["END:%s\r\n"%self.typename])
        return string.join(v, "")

class VProps:
    class String(VObject.Property):
        def get(self): return VObject.Property.get(self)
        def set(self, s): VObject.Property.set(self,s)
    class List(VObject.Property):
        def get(self):
            return string.split(VObject.Property.get(self), ";")
        def set(self, l):
            VObject.Property.set(self,string.join(l, ";"))

class _HighLevelHandler(VHandler):
    def __init__(self):
        self.root_objects=[]
        self.object_stack = []
    def handle_BEGIN(self, group, params, value):
        global _ObjectTypes
        if self.object_stack: parent = self.object_stack[0]
        else: parent = None
        try:
            object = _ObjectTypes[value](parent)
        except:
            object = VObject(parent, value)
        self.object_stack.insert(0, object)
    def handle_END(self, group, params, value):
        if self.object_stack:
            object = self.object_stack[0]
            if object.typename != value:
                print "warning: type mismatch in END:", value, object.typename
            del self.object_stack[0]
            if self.object_stack:
            	self.object_stack[0].handle_subobject(object)
            else:
                self.root_objects.append(object)
    def handle_eof(self):
        while self.object_stack:
            self.handle_END({}, [], self.object_stack[0].typename)
            del self.object_stack[0]
    def handle_unknown(self, name, group, params, value):
        object = self.object_stack[0]
        _handle(object, name, (group, params, value))

def VReadObjects(stream):
    """ VReadObjects(stream) -> VObjects
    Reads VObjects from a stream and returns them in a list
    """
    
    handler = _HighLevelHandler()
    reader=VReader(stream, handler)
    return handler.root_objects

def _VRegisterHandler(handlerclass):
    global _ObjectTypes
    _ObjectTypes[handlerclass.typename] = handlerclass

#
# -=- MAIN routine for testing
#

if __name__ == "__main__":
    f=open("D:\\james\\Python\\Jen Yee Chan.txt")
    print repr(VReadObjects(f)[0])
